/////////////////////////////// Point Multiplication ///////////////////////////////////
module ECPM(clk, rst_n, in_valid, k, I_gx, I_gy, O_rdy, O_x, O_y);

input  wire clk;
input  wire rst_n;
input  wire in_valid;
input  wire [255:0] k;
input  wire	[255:0]	I_gx;
input  wire	[255:0]	I_gy;
output wire [255:0] O_x;
output wire [255:0] O_y;
output wire O_rdy;

reg    [255:0]  k_reg;
reg    [255:0]  x;
reg    [255:0]  y;
reg             out_valid;

reg    [7:0]    cnt;
reg    [2:0]    cs_ECPM,ns_ECPM;

reg    [255:0]  InvOut_reg;
reg    [255:0]  PD_inx,PD_iny,PD_inz,PA_inx1,PA_inx2,PA_iny1,PA_iny2,PA_inz1,PA_inz2,Inv_ina,Radix4_inx,Radix4_iny;
reg             PD_invalid,PA_invalid,Inv_invalid,Radix4_invalid;
wire   [255:0]  PD_outx,PD_outy,PD_outz,PA_outx,PA_outy,PA_outz,Inv_out,Radix4_outz;
wire            PD_outvalid,PA_outvalid,Inv_outvalid,Radix4_outvalid;

parameter   Idle_ECPM    = 3'd0;  // input the privite key
parameter   Shift_ECPM   = 3'd1;  // calculate how many time to do ECPM
parameter   ML_ECPMa     = 3'd2;  // calculate the point projective coordinate
parameter   ML_ECPMb     = 3'd3;  // calculate the point projective coordinate
parameter   Inv_ECPM     = 3'd4;  // calculate inversion of Z
parameter   First_Mult   = 3'd5;  // calculate O_x 
parameter   Sec_Mult     = 3'd6;  // calculate O_y
parameter   Final_output = 3'd7;  // out_valid

PDouble     PD_module(.X(PD_inx),.Y(PD_iny),.Z(PD_inz),.in_valid(PD_invalid),.clk(clk),.rst_n(rst_n),.X2(PD_outx),.Y2(PD_outy),.Z2(PD_outz),.out_valid(PD_outvalid));
PAddition   PA_module(.X1(PA_inx1),.Y1(PA_iny1),.Z1(PA_inz1),.X2(PA_inx2),.Y2(PA_iny2),.Z2(PA_inz2),.in_valid(PA_invalid),.clk(clk),.rst_n(rst_n),.X3(PA_outx),.Y3(PA_outy),.Z3(PA_outz),.out_valid(PA_outvalid));
InvFp      Inv_module(.A(Inv_ina),.in_valid(Inv_invalid),.clk(clk),.rst_n(rst_n),.A_inv(Inv_out),.out_valid(Inv_outvalid));
radix4    mult_module(.X(Radix4_inx),.Y(Radix4_iny),.Z(Radix4_outz),.clk(clk),.rst_n(rst_n),.in_valid(Radix4_invalid),.out_valid(Radix4_outvalid));

always @(posedge clk) begin  // ECPMcs block (Sequential) (Current_state)
  if (!rst_n) cs_ECPM <= Idle_ECPM;
  else        cs_ECPM <= ns_ECPM;
end

always @(*)begin  ///////////// ECPMns block (Combinational) (Next_state_state)
  case(cs_ECPM)
    Idle_ECPM:     ns_ECPM = (in_valid == 1'b1)? Shift_ECPM : Idle_ECPM;
	Shift_ECPM:    ns_ECPM = (k_reg[cnt] == 1'b1)? ML_ECPMa : Shift_ECPM;
	ML_ECPMa:      ns_ECPM = ML_ECPMb;
	ML_ECPMb:      begin
	               if(PA_outvalid == 1'b1) begin
	                 if(cnt == 0) ns_ECPM = Inv_ECPM;
	                 else         ns_ECPM = ML_ECPMa;
	               end
	               else begin
	                 ns_ECPM = ML_ECPMb;
	               end
	end
	Inv_ECPM:      ns_ECPM = (Inv_outvalid == 1'b1)? First_Mult: Inv_ECPM;
	First_Mult:    ns_ECPM = (Radix4_outvalid == 1'b1)? Sec_Mult: First_Mult;
	Sec_Mult:      ns_ECPM = (Radix4_outvalid == 1'b1)? Final_output: Sec_Mult;
	Final_output:  ns_ECPM = Final_output;
  endcase
end

always @(posedge clk) begin  // all register in the system
  if (!rst_n) begin
    InvOut_reg <= 256'b0;
    k_reg <= 256'b0;
	x <= 256'b0;
	y <= 256'b0;
	out_valid <= 1'b0;
	cnt <= 8'd255;
	cs_ECPM <= 3'b0;
	ns_ECPM <= 3'b0;
	PD_inx <= 256'b0;
	PD_iny <= 256'b0;
	PD_inz <= 256'b0;
	PA_inx1 <= I_gx;/////I_gx initial 
	PA_inx2 <= 256'h31241DDB9A7C254AEA224B87B7B0F909886EC1DDFA71625B7ABA864C18300A57;
	PA_iny1 <= I_gy;/////I_gy initial
	PA_iny2 <= 256'h3324984C6CC933DB69B782FC3AC951F60A47AA662BBE321C924B2CD95E2D7FD7;
	PA_inz1 <= 256'h1;
	PA_inz2 <= 256'h6986C5796B577C574098D1FA3B426292EBD36100339299D16374A93D0278DE6B;
	Inv_ina <= 256'b0;
	Radix4_inx <= 256'b0;
	Radix4_iny <= 256'b0;
	PD_invalid <= 1'b0;
	PA_invalid <= 1'b0;
	Inv_invalid <= 1'b0;
	Radix4_invalid <= 1'b0;
  end
  else begin
    case(cs_ECPM)
		Idle_ECPM: begin
			if(in_valid == 1'b1) begin
				k_reg <= k;
				cnt <= 8'd255;
			end
			else begin
				k_reg <= 256'b0;
				InvOut_reg <= 256'b0;
				out_valid <= 1'b0;
				cnt <= 8'd255;
				cs_ECPM <= 3'b0;
				ns_ECPM <= 3'b0;
				PD_inx <= 256'b0;
				PD_iny <= 256'b0;
				PD_inz <= 256'b0;
				PA_inx1 <= 256'h216936D3CD6E53FEC0A4E231FDD6DC5C692CC7609525A7B2C9562D608F25D51A;
				PA_inx2 <= 256'h31241DDB9A7C254AEA224B87B7B0F909886EC1DDFA71625B7ABA864C18300A57;
				PA_iny1 <= 256'h6666666666666666666666666666666666666666666666666666666666666658;
				PA_iny2 <= 256'h3324984C6CC933DB69B782FC3AC951F60A47AA662BBE321C924B2CD95E2D7FD7;
				PA_inz1 <= 256'h1;
				PA_inz2 <= 256'h6986C5796B577C574098D1FA3B426292EBD36100339299D16374A93D0278DE6B;
				Inv_ina <= 256'b0;
				Radix4_inx <= 256'b0;
				Radix4_iny <= 256'b0;
				PD_invalid <= 1'b0;
				PA_invalid <= 1'b0;
				Inv_invalid <= 1'b0;
				Radix4_invalid <= 1'b0;
			end
		end
		Shift_ECPM: begin
			if ( (k_reg[cnt] == 1'b1) && ((k_reg[cnt-1]) == 1'b1) ) begin
				PA_invalid <= 1'b1;
				PD_invalid <= 1'b1;
				PD_inx <= 256'h31241DDB9A7C254AEA224B87B7B0F909886EC1DDFA71625B7ABA864C18300A57;
				PD_iny <= 256'h3324984C6CC933DB69B782FC3AC951F60A47AA662BBE321C924B2CD95E2D7FD7;
				PD_inz <= 256'h6986C5796B577C574098D1FA3B426292EBD36100339299D16374A93D0278DE6B;
			end
			if ( (k_reg[cnt] == 1'b1) && ((k_reg[cnt-1]) == 1'b0) ) begin
				PA_invalid <= 1'b1;
				PD_invalid <= 1'b1;
				PD_inx <= 256'h216936D3CD6E53FEC0A4E231FDD6DC5C692CC7609525A7B2C9562D608F25D51A;
				PD_iny <= 256'h6666666666666666666666666666666666666666666666666666666666666658;
				PD_inz <= 256'h1;
			end
			cnt <= cnt - 1'b1;
		end
		ML_ECPMa: begin
			PA_invalid <= 1'b1;
			PD_invalid <= 1'b1;
		end
		ML_ECPMb: begin
			PA_invalid <= 1'b0;
			PD_invalid <= 1'b0;
			if ((PA_outvalid == 1'b1) && (k_reg[cnt] == 1'b1)) begin
				PA_inx1 <= PA_outx;
				PA_iny1 <= PA_outy;
				PA_inz1 <= PA_outz;
				PA_inx2 <= PD_outx;
				PA_iny2 <= PD_outy;
				PA_inz2 <= PD_outz;
				if(k_reg[cnt-1] == 1'b1) begin
					PD_inx <= PD_outx;
					PD_iny <= PD_outy;
					PD_inz <= PD_outz;
				end
				else begin
					PD_inx <= PA_outx;
					PD_iny <= PA_outy;
					PD_inz <= PA_outz;
				end
			end
			if ((PA_outvalid == 1'b1) && (k_reg[cnt] == 1'b0)) begin//////////////need to fix (latch)
				PA_inx2 <= PA_outx;
				PA_iny2 <= PA_outy;
				PA_inz2 <= PA_outz;
				PA_inx1 <= PD_outx;
				PA_iny1 <= PD_outy;
				PA_inz1 <= PD_outz;
				if(k_reg[cnt-1] == 1'b1) begin
					PD_inx <= PA_outx;
					PD_iny <= PA_outy;
					PD_inz <= PA_outz;
				end
				else begin
					PD_inx <= PD_outx;
					PD_iny <= PD_outy;
					PD_inz <= PD_outz;
				end
			end/////////////////////////here
		if ( (PA_outvalid == 1'b1) && (cnt != 1'b0) ) begin
		  cnt <= cnt - 1'b1;
		end
		if ( (PA_outvalid == 1'b1) && (cnt == 1'b0) ) begin
		  PA_invalid <= 1'b0;
		  PD_invalid <= 1'b0;
		end
	  end
	  Inv_ECPM: begin
	    if (Inv_outvalid == 1'b1) begin
		  Inv_invalid <= 1'b0;
		  InvOut_reg <= Inv_out;
		  Radix4_inx <= PA_inx1;
		  Radix4_iny <= Inv_out;
		  Radix4_invalid <= 1'b1;
		end
		else begin
		  Inv_ina <= PA_inz1;
		  Inv_invalid <= 1'b1;
		end
	  end
	  First_Mult: begin
	    if(Radix4_outvalid == 1'b1) begin
		  Radix4_inx <= PA_iny1;
		  Radix4_iny <= InvOut_reg;
		  x <= Radix4_outz;               
		  Radix4_invalid <= 1'b1;
		  end
		  else begin
			Radix4_invalid <= 1'b0;
		  end
	  end
	  Sec_Mult: begin
	    if(Radix4_outvalid == 1'b1) begin
		  y <= Radix4_outz;
		  out_valid <= 1'b1;
		end
	  end
	  Final_output: begin
	    out_valid <= 1'b1;
	  end
	endcase
  end
end
//
		// Output Mapping
		//

	assign	O_x = x;
	assign	O_y = y;
		//
		// Ready Flag Logic
		//
	reg rdy_reg = 1'b1;
	assign O_rdy = rdy_reg;
		
	always @(posedge clk or negedge rst_n)
		
		if (rst_n == 1'b0) rdy_reg <= 1'b1;
		else begin
			
				/* clear flag */
			if ((cs_ECPM == Idle_ECPM) && in_valid)
				rdy_reg <= 1'b0;
			
				/* set flag */
			if (cs_ECPM == Final_output)
				rdy_reg <= 1'b1;
				
		end




endmodule

///////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////// Point Double ///////////////////////////////////
module PDouble(X,Y,Z,in_valid,clk,rst_n,X2,Y2,Z2,out_valid);

input   [255:0]  X,Y,Z;
input            in_valid,clk,rst_n;
output  [255:0]  X2,Y2,Z2;
output           out_valid;

parameter  PD_Idle = 5'd0;
parameter  S1a     = 5'd1;
parameter  S1b     = 5'd2;
parameter  S2a     = 5'd3;
parameter  S2b     = 5'd4;
parameter  S3a     = 5'd5;
parameter  S3b     = 5'd6;
parameter  S3c     = 5'd7;
parameter  S4a     = 5'd8;
parameter  S4b     = 5'd9;
parameter  S5a     = 5'd10;
parameter  S5b     = 5'd11;
parameter  S6      = 5'd12;
parameter  S7a     = 5'd13;
parameter  S7b     = 5'd14;
parameter  S8a     = 5'd15;
parameter  S8b     = 5'd16;
parameter  PD_Out  = 5'd17;

reg     [4:0]    dbl_cs,dbl_ns;
reg     [255:0]  B,C,D,E,F,H,J,X_reg,Y_reg,Z_reg;  // register array for point doubling
reg     [255:0]  As_ina,As_inb;  // register for add/sub unit input
reg              As_ins;  ///////// register for add/sub unit input
reg     [255:0]  Mult_inx,Mult_iny;  // register for multiplier input
reg              Mult_invalid; //////// register for multiplier input
reg     [255:0]  X2,Y2,Z2;  /////////// output register
reg              out_valid;
wire    [255:0]  As_outy;   /////////// register for add/sub unit output
wire    [255:0]  Mult_outz; /////////// register for multiplier output
wire             Mult_outvalid; /////// register for multiplier output

modaddsub  addsubUnit(.A(As_ina), .B(As_inb), .s(As_ins), .Y(As_outy));
radix4     multUnit(.X(Mult_inx), .Y(Mult_iny), .Z(Mult_outz), .clk(clk), .rst_n(rst_n), .in_valid(Mult_invalid), .out_valid(Mult_outvalid));

always @(posedge clk) begin  // PDcs block (Sequential) (Current_state)
  if (!rst_n) dbl_cs <= PD_Idle;
  else dbl_cs <= dbl_ns;
end

always @(*) begin ///////////// PDns block (Combinational)  (Next_state)
  case(dbl_cs)
    PD_Idle : dbl_ns = (in_valid == 1'b1)? S1a: PD_Idle;
    S1a     : dbl_ns = S1b;
    S1b     : dbl_ns = (Mult_outvalid == 1'b1)? S2a: S1b;
    S2a     : dbl_ns = S2b;
    S2b     : dbl_ns = (Mult_outvalid == 1'b1)? S3a: S2b;
    S3a     : dbl_ns = S3b;
    S3b     : dbl_ns = S3c;
    S3c     : dbl_ns = (Mult_outvalid == 1'b1)? S4a: S3c;
    S4a     : dbl_ns = S4b;
    S4b     : dbl_ns = (Mult_outvalid == 1'b1)? S5a: S4b;
    S5a     : dbl_ns = S5b;
    S5b     : dbl_ns = (Mult_outvalid == 1'b1)? S6: S5b;
    S6      : dbl_ns = S7a;
    S7a     : dbl_ns = S7b;
    S7b     : dbl_ns = (Mult_outvalid == 1'b1)? S8a: S7b;
    S8a     : dbl_ns = S8b;
    S8b     : dbl_ns = (Mult_outvalid == 1'b1)? PD_Out: S8b;
    PD_Out  : dbl_ns = PD_Idle;
  endcase
end

always @(posedge clk) begin
  if(!rst_n) begin
    B <= 256'b0;
	C <= 256'b0;
	D <= 256'b0;
	E <= 256'b0;
	F <= 256'b0;
	H <= 256'b0;
	J <= 256'b0;
	As_ina <= 256'b0;
	As_inb <= 256'b0;
	As_ins <= 1'b0;
	Mult_inx <= 256'b0;
	Mult_iny <= 256'b0;
	Mult_invalid <= 1'b0;
	X2 <= 256'b0;
	Y2 <= 256'b0;
    Z2 <= 256'b0;
	out_valid <= 1'b0;
	X_reg <= 256'b0;
	Y_reg <= 256'b0;
	Z_reg <= 256'b0;
  end
  else begin
    case(dbl_cs)
	  PD_Idle: begin
	    B <= 256'b0;
	    C <= 256'b0;
	    D <= 256'b0;
	    E <= 256'b0;
	    F <= 256'b0;
	    H <= 256'b0;
	    J <= 256'b0;
	    out_valid <= 1'b0;
        if(in_valid == 1'b1) begin
		  X_reg <= X;
		  Y_reg <= Y;
		  Z_reg <= Z;
		  Mult_inx <= Y;
		  Mult_iny <= Y;
		  Mult_invalid <= 1'b1;
          As_ina <= X;
		  As_inb <= Y;
		  As_ins <= 1'b0;
		end
		else begin
		  X_reg <= 256'b0;
		  Y_reg <= 256'b0;
		  Z_reg <= 256'b0;
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
          As_ina <= 256'b0;
		  As_inb <= 256'b0;
		  As_ins <= 1'b0;
		end
	  end
	  S1a: begin
	    X_reg <= As_ina;
		Mult_invalid <=  1'b1;
		B <= As_outy;
	  end
	  S1b: begin
	    if(Mult_outvalid == 1'b1) begin
		  D <= Mult_outz;
		  Mult_inx <= X_reg;
		  Mult_iny <= X_reg;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S2a: begin
	    Mult_invalid <= 1'b1;
	  end
	  S2b: begin
	    if(Mult_outvalid == 1'b1) begin
		  C <= Mult_outz;
		  Mult_inx <= B;
		  Mult_iny <= B;
		  Mult_invalid <= 1'b1;
		  As_ina <= 256'b0;
		  As_inb <= Mult_outz;
		  As_ins <= 1'b1;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S3a: begin
	    Mult_invalid <= 1'b1;
		E <= As_outy;
		As_ina <= D;
		As_inb <= C;
		As_ins <= 1'b1;
	  end
	  S3b: begin
	    Mult_invalid <= 1'b0;
	    F <= As_outy;
		As_ina <= E;
		As_inb <= D;
		As_ins <= 1'b1;
	  end
	  S3c: begin
	    if(Mult_outvalid == 1'b1) begin
		  B <= Mult_outz;
		  Mult_inx <= Z_reg;
		  Mult_iny <= Z_reg;
		  Mult_invalid <= 1'b1;
		  Y2 <= As_outy;
		  As_ina <= Mult_outz;
		  As_inb <= C;
		  As_ins <= 1'b1;
		end
		else begin
		  Mult_invalid <= 1'b0;
		end
	  end
	  S4a: begin
	    X2 <= As_outy;
		Mult_invalid <= 1'b1;
		As_ina <= As_outy;
		As_inb <= D;
		As_ins <= 1'b1;
	  end
	  S4b: begin
	    if(Mult_outvalid == 1'b1) begin
		  X2<= As_outy;
		  H <= Mult_outz;
		  Mult_inx <= F;
		  Mult_iny <= Y2;
		  Mult_invalid <= 1'b1;
		  As_ina <= Mult_outz;
		  As_inb <= Mult_outz;
		  As_ins <= 1'b0;
		end
		else begin
		  Mult_invalid <= 1'b0;
		end
	  end
	  S5a: begin
	    Mult_invalid <= 1'b1;
		J <= As_outy;
	  end
	  S5b: begin
	    if(Mult_outvalid == 1'b1) begin
		  Y2 <= Mult_outz;
		  Mult_inx <= J;
		  Mult_iny <= X2;
		  As_ina <= F;
		  As_inb <= J;
		  As_ins <= 1'b1;
		  Mult_invalid <= 1'b0;
		end
		else begin
		  Mult_invalid <= 1'b0;
		end
	  end
	  S6: begin
	    J <= As_outy;
	    Mult_inx <= As_outy;
		Mult_invalid <= 1'b1;
	  end
	  S7a: begin
	    Mult_invalid <= 1'b1;
	  end
	  S7b: begin
	    if(Mult_outvalid == 1'b1) begin
		  X2 <= Mult_outz;
		  Mult_inx <= F;
		  Mult_iny <= J;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_invalid <= 1'b0;
		end
	  end
	  S8a: begin
	    Mult_invalid <= 1'b1;
	  end
	  S8b: begin
	    if(Mult_outvalid == 1'b1) begin
		  Z2 <= Mult_outz;
		  out_valid <= 1'b1;
		end
		else begin
		  Mult_invalid <= 1'b0;
		end
	  end
	  PD_Out: begin
	    out_valid <= 1'b0;
	  end
	endcase
  end
end

endmodule
////////////////////////////////////////////////////////////////////////////////

/////////////////////////////// Point Addition ///////////////////////////////////
module PAddition(X1,Y1,Z1,X2,Y2,Z2,in_valid,clk,rst_n,X3,Y3,Z3,out_valid);

input   [255:0]  X1,Y1,Z1,X2,Y2,Z2;
input            in_valid,clk,rst_n;
output  [255:0]  X3,Y3,Z3;
output           out_valid;

parameter  PA_Idle = 5'd0;
parameter  S1a     = 5'd1;
parameter  S1b     = 5'd2;
parameter  S2a     = 5'd3;
parameter  S2b     = 5'd4;
parameter  S3a     = 5'd5;
parameter  S3b     = 5'd6;
parameter  S4a     = 5'd7;
parameter  S4b     = 5'd8;
parameter  S5a     = 5'd9;
parameter  S5b     = 5'd10;
parameter  S6a     = 5'd11;
parameter  S6b     = 5'd12;
parameter  S7a     = 5'd13;
parameter  S7b     = 5'd14;
parameter  S8a     = 5'd15;
parameter  S8b     = 5'd16;
parameter  S9a     = 5'd17;
parameter  S9b     = 5'd18;
parameter  S10a    = 5'd19;
parameter  S10b    = 5'd20;
parameter  S11a    = 5'd21;
parameter  S11b    = 5'd22;
parameter  S12a    = 5'd23;
parameter  S12b    = 5'd24;
parameter  PA_Out  = 5'd25;

reg     [4:0]    add_cs,add_ns;
reg     [255:0]  A,B,C,D,E,F,G,r0,r1,r2,X1_reg,Y1_reg,Z1_reg,X2_reg,Y2_reg,Z2_reg;  // register array for point addition
reg     [255:0]  As_ina,As_inb;  // register for add/sub unit input
reg              As_ins;  ///////// register for add/sub unit input
reg     [255:0]  Mult_inx,Mult_iny;  // register for multiplier input
reg              Mult_invalid; //////// register for multiplier input
reg     [255:0]  X3,Y3,Z3;  /////////// output register
reg              out_valid;
wire    [255:0]  As_outy;   /////////// register for add/sub unit output
wire    [255:0]  Mult_outz; /////////// register for multiplier output
wire             Mult_outvalid; /////// register for multiplier output

modaddsub  addsubUnit2(.A(As_ina), .B(As_inb), .s(As_ins), .Y(As_outy));
radix4     multUnit2(.X(Mult_inx), .Y(Mult_iny), .Z(Mult_outz), .clk(clk), .rst_n(rst_n), .in_valid(Mult_invalid), .out_valid(Mult_outvalid));

always @(posedge clk) begin  // PAcs block (Sequential) (Current_state)
  if (!rst_n) add_cs <= PA_Idle;
  else add_cs <= add_ns;
end

always @(*) begin ///////////// PAns block (Combinational)  (Next_state)
  case(add_cs)
    PA_Idle : add_ns = (in_valid == 1'b1)? S1a: PA_Idle;
    S1a     : add_ns = S1b;
    S1b     : add_ns = (Mult_outvalid == 1'b1)? S2a: S1b;
    S2a     : add_ns = S2b;
    S2b     : add_ns = (Mult_outvalid == 1'b1)? S3a: S2b;
    S3a     : add_ns = S3b;
    S3b     : add_ns = (Mult_outvalid == 1'b1)? S4a: S3b;
    S4a     : add_ns = S4b;
    S4b     : add_ns = (Mult_outvalid == 1'b1)? S5a: S4b;
    S5a     : add_ns = S5b;
    S5b     : add_ns = (Mult_outvalid == 1'b1)? S6a: S5b;
    S6a     : add_ns = S6b;
	S6b     : add_ns = (Mult_outvalid == 1'b1)? S7a: S6b;
    S7a     : add_ns = S7b;
    S7b     : add_ns = (Mult_outvalid == 1'b1)? S8a: S7b;
    S8a     : add_ns = S8b;
    S8b     : add_ns = (Mult_outvalid == 1'b1)? S9a: S8b;
	S9a     : add_ns = S9b; 
	S9b     : add_ns = (Mult_outvalid == 1'b1)? S10a: S9b;
	S10a    : add_ns = S10b;
	S10b    : add_ns = (Mult_outvalid == 1'b1)? S11a: S10b;
	S11a    : add_ns = S11b;
	S11b    : add_ns = (Mult_outvalid == 1'b1)? S12a: S11b;
	S12a    : add_ns = S12b;
	S12b    : add_ns = (Mult_outvalid == 1'b1)? PA_Out: S12b;
    PA_Out  : add_ns = PA_Idle;
  endcase
end

always @(posedge clk) begin
  if(!rst_n) begin
    A <= 256'b0;
    B <= 256'b0;
	C <= 256'b0;
	D <= 256'b0;
	E <= 256'b0;
	F <= 256'b0;
	G <= 256'b0;
	r0 <= 256'b0;
	r1 <= 256'b0;
	r2 <= 256'b0;
	As_ina <= 256'b0;
	As_inb <= 256'b0;
	As_ins <= 1'b0;
	Mult_inx <= 256'b0;
	Mult_iny <= 256'b0;
	Mult_invalid <= 1'b0;
	X3 <= 256'b0;
	Y3 <= 256'b0;
    Z3 <= 256'b0;
	out_valid <= 1'b0;
	X1_reg <= 256'b0;
	Y1_reg <= 256'b0;
	Z1_reg <= 256'b0;
	X2_reg <= 256'b0;
	Y2_reg <= 256'b0;
	Z2_reg <= 256'b0;
  end
  else begin
    case(add_cs)
	  PA_Idle: begin
	    A <= 256'b0;
        B <= 256'b0;
	    C <= 256'b0;
	    D <= 256'b0;
	    E <= 256'b0;
	    F <= 256'b0;
	    G <= 256'b0;
	    r0 <= 256'b0;
	    r1 <= 256'b0;
	    r2 <= 256'b0;
	    out_valid <= 1'b0;
        if(in_valid == 1'b1) begin
		  X1_reg <= X1;
		  Y1_reg <= Y1;
		  Z1_reg <= Z1;
		  X2_reg <= X2;
		  Y2_reg <= Y2;
		  Z2_reg <= Z2;
		  Mult_inx <= Z1;
		  Mult_iny <= Z2;
		  Mult_invalid <= 1'b1;
          As_ina <= X1;
		  As_inb <= Y1;
		  As_ins <= 1'b0;
		end
		else begin
		  X1_reg <= 256'b0;
		  Y1_reg <= 256'b0;
		  Z1_reg <= 256'b0;
		  X2_reg <= 256'b0;
		  Y2_reg <= 256'b0;
		  Z2_reg <= 256'b0;
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
          As_ina <= 256'b0;
		  As_inb <= 256'b0;
		  As_ins <= 1'b0;
		end
	  end
	  S1a: begin
		Mult_invalid <=  1'b1;
		X3 <= As_outy;
		As_ina <= X2_reg;
		As_inb <= Y2_reg;
		As_ins <= 1'b0;
	  end
	  S1b: begin
	    if(Mult_outvalid == 1'b1) begin
		  A <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		  r0 <= As_outy;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S2a: begin
	    Mult_invalid <= 1'b1;
	  end
	  S2b: begin
	    if(Mult_outvalid == 1'b1) begin
		  B <= Mult_outz;
		  Mult_inx <= X1_reg;
		  Mult_iny <= X2_reg;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S3a: begin
	    Mult_invalid <= 1'b1;
	  end
	  S3b: begin
	    if(Mult_outvalid == 1'b1) begin
		  C <= Mult_outz;
		  Mult_inx <= Y1_reg;
		  Mult_iny <= Y2_reg;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S4a: begin
		Mult_invalid <= 1'b1;
	  end
	  S4b: begin
	    if(Mult_outvalid == 1'b1) begin
		  D <= Mult_outz;
		  Mult_inx <= C;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		  As_ina <= Mult_outz;
		  As_inb <= C;
		  As_ins <= 1'b0;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S5a: begin
	    Mult_invalid <= 1'b1;
		r1 <= As_outy;
	  end
	  S5b: begin
	    if(Mult_outvalid == 1'b1) begin
		  E <= Mult_outz;
		  Mult_inx <= 256'h52036CEE2B6FFE738CC740797779E89800700A4D4141D8AB75EB4DCA135978A3;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_invalid <= 1'b0;
		end
	  end
	  S6a: begin
	    Mult_invalid <= 1'b1;
	  end
	  S6b: begin
	    if(Mult_outvalid == 1'b1) begin
		  E <= Mult_outz;
		  Mult_inx <= X3;
		  Mult_iny <= r0;
		  Mult_invalid <= 1'b1;
		  As_ina <= B;
		  As_inb <= Mult_outz;
		  As_ins <= 1'b0;
		end
		else begin
		  Mult_invalid <= 1'b0;
		end
	  end
	  S7a: begin
	    Mult_invalid <= 1'b1;
		G <= As_outy;
		As_ina <= B;
		As_inb <= E;
		As_ins <= 1'b1;
	  end
	  S7b: begin
	    if(Mult_outvalid == 1'b1) begin
		  X3 <= Mult_outz;
		  Mult_inx <= A;
		  Mult_iny <= As_outy;
		  Mult_invalid <= 1'b1;
		  F <= As_outy;
		  As_ina <= Mult_outz;
		  As_inb <= C;
		  As_ins <= 1'b1;
		end
		else begin
		  Mult_invalid <= 1'b0;
		end
	  end
	  S8a: begin
	    Mult_invalid <= 1'b1;
		X3 <= As_outy;
		As_ina <= As_outy;
		As_inb <= D;
		As_ins <= 1'b1;
	  end
	  S8b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r2 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= As_outy;
		  Mult_invalid <= 1'b1;
		  X3 <= As_outy;
		end
		else begin
		  Mult_invalid <= 1'b0;
		end
	  end
	  S9a: begin
	    Mult_invalid <= 1'b1;
	  end
	  S9b: begin
	    if(Mult_outvalid == 1'b1) begin
		  X3 <= Mult_outz;
		  Mult_inx <= A;
		  Mult_iny <= G;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S10a: begin
	    Mult_invalid <= 1'b1;
	  end
	  S10b: begin
	    if(Mult_outvalid == 1'b1) begin
		  Y3 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= r1;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S11a: begin
	    Mult_invalid <= 1'b1;
	  end
	  S11b: begin
	    if(Mult_outvalid == 1'b1) begin
		  Y3 <= Mult_outz;
		  Mult_inx <= F;
		  Mult_iny <= G;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S12a: begin
	    Mult_invalid <= 1'b1;
	  end
	  S12b: begin
	    if(Mult_outvalid == 1'b1) begin
		  Z3 <= Mult_outz;
		  out_valid <= 1'b1;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	 PA_Out: begin
	    out_valid <= 1'b0;
	  end
	endcase
  end
end

endmodule
////////////////////////////////////////////////////////////////////////////////

/////////////////////////////// MulInv Fp //////////////////////////////////////
module InvFp(A,in_valid,clk,rst_n,A_inv,out_valid);

input  [255:0]  A;
input           clk,in_valid,rst_n;
output [255:0]  A_inv;
output          out_valid;

parameter   InvIdle = 6'd0;
parameter   S1a     = 6'd1;
parameter   S1b     = 6'd2;
parameter   S2a     = 6'd3;
parameter   S2b     = 6'd4;
parameter   S3a     = 6'd5;
parameter   S3b     = 6'd6;
parameter   S4a     = 6'd7;
parameter   S4b     = 6'd8;
parameter   S5a     = 6'd9;
parameter   S5b     = 6'd10;
parameter   S6a     = 6'd11;
parameter   S6b     = 6'd12;
parameter   S7a     = 6'd13;
parameter   S7b     = 6'd14;
parameter   S8a     = 6'd15;
parameter   S8b     = 6'd16;
parameter   S9a     = 6'd17;
parameter   S9b     = 6'd18;
parameter   S10a    = 6'd19;
parameter   S10b    = 6'd20;
parameter   S11a    = 6'd21;
parameter   S11b    = 6'd22;
parameter   S12a    = 6'd23;
parameter   S12b    = 6'd24;
parameter   S13a    = 6'd25;
parameter   S13b    = 6'd26;
parameter   S14a    = 6'd27;
parameter   S14b    = 6'd28;
parameter   S15a    = 6'd29;
parameter   S15b    = 6'd30;
parameter   S16a    = 6'd31;
parameter   S16b    = 6'd32;
parameter   S17a    = 6'd33;
parameter   S17b    = 6'd34;
parameter   S18a    = 6'd35;
parameter   S18b    = 6'd36;
parameter   S19a    = 6'd37;
parameter   S19b    = 6'd38;
parameter   S20a    = 6'd39;
parameter   S20b    = 6'd40;
parameter   S21a    = 6'd41;
parameter   S21b    = 6'd42;
parameter   S22a    = 6'd43;
parameter   S22b    = 6'd44;
parameter   InvOut  = 6'd45;

reg    [6:0]    InvCnt;
reg    [5:0]    InvCs,InvNs;
reg             out_valid;
reg    [255:0]  A_inv;
reg    [255:0]  Mult_inx,Mult_iny; // register for multiplier
reg             Mult_invalid;      // register for multiplier
reg    [255:0]  r1,r2,r3;  // register array
wire   [255:0]  Mult_outz;  //  output connection for multiplier
wire            Mult_outvalid;  // output connection for multiplier

radix4  mult4inv (.X(Mult_inx), .Y(Mult_iny), .Z(Mult_outz), .clk(clk), .rst_n(rst_n), .in_valid(Mult_invalid), .out_valid(Mult_outvalid));

always @(posedge clk) begin  // InvCs block (Sequential) (Current_state)
  if (!rst_n) InvCs <= InvIdle;
  else InvCs <= InvNs;
end

always @(*) begin  // InvNs block (Combinational)  (Next_state)
  case(InvCs)
    InvIdle : InvNs = (in_valid == 1'b1)? S1a: InvIdle;
    S1a     : InvNs = S1b;
    S1b     : InvNs = (Mult_outvalid == 1'b1)? S2a: S1b;
    S2a     : InvNs = S2b;
    S2b     : begin
	            if(Mult_outvalid == 1'b0) InvNs = S2b;
				else begin
				  if(InvCnt == 7'd2) InvNs = S3a;
				  else InvNs = S2a;
				end
	          end
    S3a     : InvNs = S3b;
    S3b     : InvNs = (Mult_outvalid == 1'b1)? S4a: S3b;
    S4a     : InvNs = S4b;
    S4b     : InvNs = (Mult_outvalid == 1'b1)? S5a: S4b;
    S5a     : InvNs = S5b;
    S5b     : InvNs = (Mult_outvalid == 1'b1)? S6a: S5b;
    S6a     : InvNs = S6b;
    S6b     : InvNs = (Mult_outvalid == 1'b1)? S7a: S6b;
    S7a     : InvNs = S7b;
    S7b     : begin
	            if(Mult_outvalid == 1'b0) InvNs = S7b;
				else begin
				  if(InvCnt == 7'd5) InvNs = S8a;
				  else InvNs = S7a;
				end
	          end
	S8a     : InvNs = S8b;
	S8b     : InvNs = (Mult_outvalid == 1'b1)? S9a: S8b;
	S9a     : InvNs = S9b;
	S9b     : begin
	            if(Mult_outvalid == 1'b0) InvNs = S9b;
				else begin
				  if(InvCnt == 7'd10) InvNs = S10a;
				  else InvNs = S9a;
				end
	          end
	S10a    : InvNs = S10b;
	S10b    : InvNs = (Mult_outvalid == 1'b1)? S11a: S10b;
	S11a    : InvNs = S11b;
	S11b    : begin
	            if(Mult_outvalid == 1'b0) InvNs = S11b;
				else begin
				  if(InvCnt == 7'd20) InvNs = S12a;
				  else InvNs = S11a;
				end
	          end
	S12a    : InvNs = S12b;
	S12b    : InvNs = (Mult_outvalid == 1'b1)? S13a: S12b;
	S13a    : InvNs = S13b;
	S13b    : begin
	            if(Mult_outvalid == 1'b0) InvNs = S13b;
				else begin
				  if(InvCnt == 7'd10) InvNs = S14a;
				  else InvNs = S13a;
				end
	          end
	S14a    : InvNs = S14b;
	S14b    : InvNs = (Mult_outvalid == 1'b1)? S15a: S14b;
	S15a    : InvNs = S15b;
	S15b    : begin
	            if(Mult_outvalid == 1'b0) InvNs = S15b;
				else begin
				  if(InvCnt == 7'd50) InvNs = S16a;
				  else InvNs = S15a;
				end
	          end
	S16a    : InvNs = S16b;
	S16b    : InvNs = (Mult_outvalid == 1'b1)? S17a: S16b;
	S17a    : InvNs = S17b;
	S17b    : begin
	            if(Mult_outvalid == 1'b0) InvNs = S17b;
				else begin
				  if(InvCnt == 7'd100) InvNs = S18a;
				  else InvNs = S17a;
				end
	          end
	S18a    : InvNs = S18b;
	S18b    : InvNs = (Mult_outvalid == 1'b1)? S19a: S18b;
	S19a    : InvNs = S19b;
	S19b    : begin
	            if(Mult_outvalid == 1'b0) InvNs = S19b;
				else begin
				  if(InvCnt == 7'd50) InvNs = S20a;
				  else InvNs = S19a;
				end
	          end
	S20a    : InvNs = S20b;
	S20b    : InvNs = (Mult_outvalid == 1'b1)? S21a: S20b;
	S21a    : InvNs = S21b;
	S21b    : begin
	            if(Mult_outvalid == 1'b0) InvNs = S21b;
				else begin
				  if(InvCnt == 7'd5) InvNs = S22a;
				  else InvNs = S21a;
				end
	          end
	S22a    : InvNs = S22b;
	S22b    : InvNs = (Mult_outvalid == 1'b1)? InvOut: S22b;
    InvOut  : InvNs = InvIdle;
  endcase
end

always@ (posedge clk) begin
  if(!rst_n) begin          // reset all register in the inversion
    Mult_inx <= 256'b0;
	Mult_iny <= 256'b0;
	Mult_invalid <= 1'b0;
	A_inv <= 256'b0;
	out_valid <= 1'b0;
	r1 <= 256'b0;
	r2 <= 256'b0;
	r3 <= 256'b0;
	InvCnt <= 7'b0;
  end
  else begin
    case (InvCs)
	  InvIdle:  begin    ////////////////////////////  Inv Idle Stage
	    out_valid <= 1'b0;
		InvCnt <= 7'b0;
	    if(in_valid == 1'b1) begin 
	      r1 <= A;
	      Mult_inx <= A;
	      Mult_iny <= A;
	      Mult_invalid <= 1'b1;
	    end  
	    else begin 
	      Mult_inx <= 256'b0;
	      Mult_iny <= 256'b0;
	      Mult_invalid <= 1'b0;
	    end
	    r2 <= 256'b0;
	    r3 <= 256'b0;
	  end
	  S1a: begin        ///////////////////////////   S1
		Mult_invalid <= 1'b1;
		r1 <= Mult_inx;
	  end
	  S1b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r2 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S2a: begin        ///////////////////////////   S2
		Mult_invalid <= 1'b1;
		InvCnt <= InvCnt + 1'b1;
	  end
	  S2b: begin
	    if(Mult_outvalid == 1'b0) begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
		else begin
		  if(InvCnt == 7'd2) begin
		    InvCnt <= 7'b0;
			r3 <= Mult_outz;
			Mult_inx <=  Mult_outz;
			Mult_iny <=  r1;
			Mult_invalid <= 1'b1;
		  end
		  else begin
		    Mult_inx <= Mult_outz;
			Mult_iny <= Mult_outz;
			Mult_invalid <= 1'b1;
		  end
		end
	  end
	  S3a: begin        ///////////////////////////   S3
		Mult_invalid <= 1'b1;
	  end
	  S3b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r1 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= r2;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S4a: begin        ///////////////////////////   S4
		Mult_invalid <= 1'b1;
	  end
	  S4b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r2 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S5a: begin        ///////////////////////////   S5
		Mult_invalid <= 1'b1;
	  end
	  S5b: begin
	    if(Mult_outvalid == 1'b1) begin
		  Mult_inx <= Mult_outz;
		  Mult_iny <= r1;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S6a: begin        ///////////////////////////   S6
		Mult_invalid <= 1'b1;
	  end
	  S6b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r1 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S7a: begin        ///////////////////////////   S7
		Mult_invalid <= 1'b1;
		InvCnt <= InvCnt + 1'b1;
	  end
	  S7b: begin
	    if(Mult_outvalid == 1'b0) begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
		else begin
		  if(InvCnt == 7'd5) begin
		    InvCnt <= 7'b0;
			Mult_inx <=  Mult_outz;
			Mult_iny <=  r1;
			Mult_invalid <= 1'b1;
		  end
		  else begin
		    Mult_inx <= Mult_outz;
			Mult_iny <= Mult_outz;
			Mult_invalid <= 1'b1;
		  end
		end
	  end
	  S8a: begin        ///////////////////////////   S8
		Mult_invalid <= 1'b1;
	  end
	  S8b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r1 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S9a: begin        ///////////////////////////   S9
		Mult_invalid <= 1'b1;
		InvCnt <= InvCnt + 1'b1;
	  end
	  S9b: begin
	    if(Mult_outvalid == 1'b0) begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
		else begin
		  if(InvCnt == 7'd10) begin
		    InvCnt <= 7'b0;
			Mult_inx <=  Mult_outz;
			Mult_iny <=  r1;
			Mult_invalid <= 1'b1;
		  end
		  else begin
		    Mult_inx <= Mult_outz;
			Mult_iny <= Mult_outz;
			Mult_invalid <= 1'b1;
		  end
		end
	  end
	  S10a: begin        ///////////////////////////   S10
		Mult_invalid <= 1'b1;
	  end
	  S10b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r3 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S11a: begin        ///////////////////////////   S11
		Mult_invalid <= 1'b1;
		InvCnt <= InvCnt + 1'b1;
	  end
	  S11b: begin
	    if(Mult_outvalid == 1'b0) begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
		else begin
		  if(InvCnt == 7'd20) begin
		    InvCnt <= 7'b0;
			Mult_inx <=  Mult_outz;
			Mult_iny <=  r3;
			Mult_invalid <= 1'b1;
		  end
		  else begin
		    Mult_inx <= Mult_outz;
			Mult_iny <= Mult_outz;
			Mult_invalid <= 1'b1;
		  end
		end
	  end
	  S12a: begin        ///////////////////////////   S12
		Mult_invalid <= 1'b1;
	  end
	  S12b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r3 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S13a: begin        ///////////////////////////   S13
		Mult_invalid <= 1'b1;
		InvCnt <= InvCnt + 1'b1;
	  end
	  S13b: begin
	    if(Mult_outvalid == 1'b0) begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
		else begin
		  if(InvCnt == 7'd10) begin
		    InvCnt <= 7'b0;
			Mult_inx <=  Mult_outz;
			Mult_iny <=  r1;
			Mult_invalid <= 1'b1;
		  end
		  else begin
		    Mult_inx <= Mult_outz;
			Mult_iny <= Mult_outz;
			Mult_invalid <= 1'b1;
		  end
		end
	  end
	  S14a: begin        ///////////////////////////   S14
		Mult_invalid <= 1'b1;
	  end
	  S14b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r1 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S15a: begin        ///////////////////////////   S15
		Mult_invalid <= 1'b1;
		InvCnt <= InvCnt + 1'b1;
	  end
	  S15b: begin
	    if(Mult_outvalid == 1'b0) begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
		else begin
		  if(InvCnt == 7'd50) begin
		    InvCnt <= 7'b0;
			Mult_inx <=  Mult_outz;
			Mult_iny <=  r1;
			Mult_invalid <= 1'b1;
		  end
		  else begin
		    Mult_inx <= Mult_outz;
			Mult_iny <= Mult_outz;
			Mult_invalid <= 1'b1;
		  end
		end
	  end
	  S16a: begin        ///////////////////////////   S16
		Mult_invalid <= 1'b1;
	  end
	  S16b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r3 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S17a: begin        ///////////////////////////   S17
		Mult_invalid <= 1'b1;
		InvCnt <= InvCnt + 1'b1;
	  end
	  S17b: begin
	    if(Mult_outvalid == 1'b0) begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
		else begin
		  if(InvCnt == 7'd100) begin
		    InvCnt <= 7'b0;
			Mult_inx <=  Mult_outz;
			Mult_iny <=  r3;
			Mult_invalid <= 1'b1;
		  end
		  else begin
		    Mult_inx <= Mult_outz;
			Mult_iny <= Mult_outz;
			Mult_invalid <= 1'b1;
		  end
		end
	  end
	  S18a: begin        ///////////////////////////   S18
		Mult_invalid <= 1'b1;
	  end
	  S18b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r3 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S19a: begin        ///////////////////////////   S19
		Mult_invalid <= 1'b1;
		InvCnt <= InvCnt + 1'b1;
	  end
	  S19b: begin
	    if(Mult_outvalid == 1'b0) begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
		else begin
		  if(InvCnt == 7'd50) begin
		    InvCnt <= 7'b0;
			Mult_inx <=  Mult_outz;
			Mult_iny <=  r1;
			Mult_invalid <= 1'b1;
		  end
		  else begin
		    Mult_inx <= Mult_outz;
			Mult_iny <= Mult_outz;
			Mult_invalid <= 1'b1;
		  end
		end
	  end
	  S20a: begin        ///////////////////////////   S20
		Mult_invalid <= 1'b1;
	  end
	  S20b: begin
	    if(Mult_outvalid == 1'b1) begin
		  r1 <= Mult_outz;
		  Mult_inx <= Mult_outz;
		  Mult_iny <= Mult_outz;
		  Mult_invalid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  S21a: begin        ///////////////////////////   S21
		Mult_invalid <= 1'b1;
		InvCnt <= InvCnt + 1'b1;
	  end
	  S21b: begin
	    if(Mult_outvalid == 1'b0) begin
		  Mult_inx <= 256'b0;
		  Mult_iny <= 256'b0;
		  Mult_invalid <= 1'b0;
		end
		else begin
		  if(InvCnt == 7'd5) begin
		    InvCnt <= 7'b0;
			Mult_inx <=  Mult_outz;
			Mult_iny <=  r2;
			Mult_invalid <= 1'b1;
		  end
		  else begin
		    Mult_inx <= Mult_outz;
			Mult_iny <= Mult_outz;
			Mult_invalid <= 1'b1;
		  end
		end
	  end
	  S22a: begin        ///////////////////////////   S22
		Mult_invalid <= 1'b1;
	  end
	  S22b: begin
	    if(Mult_outvalid == 1'b1) begin
		  A_inv <= Mult_outz;
		  out_valid <= 1'b1;
		end
		else begin
		  Mult_inx <= 6'b0;
		  Mult_iny <= 6'b0;
		  Mult_invalid <= 1'b0;
		end
	  end
	  InvOut: begin
	    out_valid <= 1'b0;
	  end
	  
	endcase  
  end
end

endmodule 
////////////////////////////////////////////////////////////////////////////////

/////////////////////////////// Radix4 Fp //////////////////////////////////////
module radix4(X,Y,Z,clk,rst_n,in_valid,out_valid);

input  clk,rst_n,in_valid;
input  [255:0] X,Y; //X: multiplicand Y: multiplier
output [255:0] Z; // X? mod pirme
output out_valid;

parameter  p  = 256'h7FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFED; // prime of Edwards25519
parameter  S0 = 2'b00; // Idle stage
parameter  S1 = 2'b01; // save Y0 to Y0_shift_reg
parameter  S2 = 2'b10; // Loop stage
parameter  S3 = 2'b11; // output stage

reg              out_valid;
reg    [1:0]     current_state,next_state;
reg    [255:0]   X_in_reg;
reg    [255:0]   Z;
reg    [255:0]   W,T; //register of iteration 
reg    [7:0]     i;  //iteration count 0~127 (128 times)
reg    [256:0]   Y0_shift_reg; // Y0 shift register in iteration
reg              addsub; // addsub control signal
wire             Y0_SR; // Y0 shift register control signal
wire   [256:0]   Y0; // multiplier padded 0  
wire   [255:0]   X2; // multiplicand ?2 mod prime
wire   [255:0] W4; // iteration W ?4 mod prime
wire   [255:0] addsubout; // addsub output

assign Y0 = {Y, 1'b0};
assign Y0_SR = (current_state == 2'b10)? 1'b1: 1'b0;
doubleFp doubleFp(.W(X_in_reg), .W2(X2));
doudou   quadraFp(.W(W), .W4(W4));
modaddsub modaddsub(.A(W4), .B(T), .s(addsub), .Y(addsubout));

always @(posedge clk) begin  //current_state block
  if (!rst_n) current_state <= S0;
  else current_state <= next_state;
end

always @(current_state or i or in_valid) begin  // next_state block
  case(current_state)
    S0: next_state = (in_valid ==  1'b1)? S1: S0;
	  S1: next_state = S2;
	  S2: next_state = (i == 8'd127)? S3: S2;
	  S3: next_state = S0;
  endcase
end

always@(posedge clk) begin   // Sequential block
  if(!rst_n) begin // reset all register
    Z <= 256'b0;
	  W <= 256'b0;
	  T <= 256'b0;
	  i <= 8'b0;
	  addsub <= 1'b0;
	  Y0_shift_reg <= 257'b0;
	  X_in_reg <= 256'b0;
	  out_valid <= 1'b0;
  end
  else begin
    case(current_state)
	  S0: begin
	    W <= 256'b0;
	    T <= 256'b0;
	    i <= 8'b0;
	    addsub <= 1'b0;
	    out_valid <= 1'b0;
	    if(next_state == S1) begin
	    X_in_reg <= X;
		  Y0_shift_reg <= Y0;
		  Z <= 256'b0;
		end
		else begin 
		  Y0_shift_reg <= 257'b0;
		  X_in_reg <= 256'b0;
		end
	  end
	  S1: begin
	    Z <= 256'b0;
	    W <= 256'b0;
	    T <= 256'b0;
	    i <= 8'b0;
	    addsub <= 1'b0;
	  end
	  S2: begin
	    Z <= 256'b0;
		  W <= addsubout;
		  Y0_shift_reg <= (Y0_shift_reg << 2);
		  case(Y0_shift_reg[256:254])
		  3'b000: begin 
		    T <= 256'b0;
		    addsub <= 1'b0;
		  end
		  3'b001: begin 
		    T <= X_in_reg;
			  addsub <= 1'b0;
		  end
		  3'b010: begin 
		    T <= X_in_reg;
			addsub <= 1'b0;
		  end
		  3'b011: begin 
		    T <= X2;
			  addsub <= 1'b0;
		  end
		  3'b100: begin 
		    T <= X2;
		    addsub <= 1'b1;
		  end
		  3'b101: begin 
		    T <= X_in_reg;
			  addsub <= 1'b1;
		  end
		  3'b110: begin 
		    T <= X_in_reg;
			  addsub <= 1'b1;
		  end
		  3'b111: begin 
		    T <= 256'b0;
			  addsub <= 1'b0;
		  end
		endcase
		i <= i + 1'b1;
	  end
	  S3: begin
	    out_valid <= 1'b1;
	    Z <= addsubout;
		  W <= 256'b0;
		  i <= 8'b0;
		  T <= 256'b0;
		  Y0_shift_reg <= 257'b0;
		  addsub <= 1'b0;
	  end
	endcase
  end
end
endmodule
/////////////////////////////////////////////////////////////////////////////////

/////////////////////////////// add/sub Fp //////////////////////////////////////
module modaddsub(A,B,s,Y);
  
input [255:0] A,B;
input s;
output reg [255:0] Y;

parameter  p  = 256'h7FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFED;

reg [255:0] a,c,y1;
reg d;
reg [255:0] e,f,h,y2;
reg g;

always@(*)
begin
  a=A+B;
  {d,c}=a+~p+1;
  if(d == 1'b1) y1=a;
  else y1=c;
    
   e=~B; 
  {g,f}=A+e+1;
   h=p+f;
   if(g == 1'b1) y2=f;
   else y2=h;
     
   if(s == 1'b0) Y=y1;
   else Y=y2;
end
endmodule
////////////////////////////////////////////////////////////////////////////////

/////////////////////////////// Double Fp //////////////////////////////////////
module doubleFp(W,W2);

input  [255:0]  W;
output [255:0]  W2;

parameter  p  = 256'h7FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFED;

wire   [255:0]  w;
wire   [256:0]  a;

assign  w = W << 1;
assign  a = w + ~p + 1;
assign  W2 = (a[256])? w: a[255:0];

endmodule
////////////////////////////////////////////////////////////////////////////////

/////////////////////////////// Quadra Fp //////////////////////////////////////
module doudou(W,W4);

input  [255:0]  W;
output [255:0]  W4;

wire   [255:0]  w2;

doubleFp doubleFp001(.W(W), .W2(w2));
doubleFp doubleFp002(.W(w2), .W2(W4));

endmodule
/////////////////////////////////////////////////////////////////////////
